import numpy as np


def old_entropy_time_series(data):
    """
    Finds the entropy of a time-series assuming that it is a multi-variate
    Gaussian with one variable for each time point - a Gaussian process.

    Note that errors cannot be estimated using bootstrapping because having
    repeated samples always reduces the entropy.

    Parameters
    ----------
    data: array
        A 2D array of the data with different time points as rows
            and rows as columns

    Returns
    -------
    gaussianentropy: float
        The estimate of the entropy using the formula for a multi-variate
        Gaussian distribution
    cov: array
        The covariance matrix estimated from the data and used to calculate
        the entropy

    Example
    -------
    >>> gaussianentropy, cov = entropy_time_series(data)
    """
    # number of time points
    n = data.shape[1]
    # first find correlation matrix (the normalised covariance) because
    # it is more robust numerically
    corr = np.zeros((n, n))
    # deviation from mean over replicates at each time point for each replicate
    ddev = data - np.nanmean(data, axis=0).reshape((1, n))
    for i in np.arange(n):
        for j in np.arange(i + 1, n):
            dsi = ddev[:, i] / np.sqrt(np.nansum(ddev[:, i] ** 2))
            dsj = ddev[:, j] / np.sqrt(np.nansum(ddev[:, j] ** 2))
            corr[i, j] = np.nansum(dsi * dsj)
    corr = corr + corr.T + np.diag(np.ones(data.shape[1]))
    # # fix if poorly defined
    # if np.min(np.linalg.eig(corr)[0]) < 0:
    #     # use Higham's method
    #     # https://nhigham.com/2013/02/13/the-nearest-correlation-matrix/
    #     # implemented by Mike Croucher
    #     print("Numerical problems - using the nearest correlation matrix")
    #     corr = nearcorr(
    #         corr, max_iterations=1000, except_on_too_many_iterations=False
    #     )
    # find covariance matrix
    smat = np.diag(np.nanstd(data, axis=0))
    cov = np.matmul(smat, np.matmul(corr, smat))
    if True:
        cov = np.zeros((n, n))
        for i in np.arange(n):
            for j in np.arange(i + 1, n):
                cov[i, j] = np.nanmean(ddev[:, i] * ddev[:, j])
        cov = cov + cov.T + np.diag(np.nanvar(data, axis=0))
    # find determinant using cholesky decomposition
    if not isPD(cov):
        print("Using nearestPD")
        cov = nearestPD(cov)
    try:
        ell = np.linalg.cholesky(2 * np.pi * np.e * cov)
        gaussianentropy = np.sum(np.log(np.diag(ell)))
        return gaussianentropy, cov
    except np.linalg.LinAlgError:
        print("Error: the covariance matrix is not positive definite")
        return np.nan, np.nan


def mynearestPD(a):
    """
    From N Higham
    https://nhigham.com/2021/01/26/what-is-the-nearest-positive-semidefinite-matrix/
    """
    # a = [[0, 1, 0, 0, 0],
    #      [0, 0, 1, 0, 0],
    #      [0, 0, 0, 1, 0],
    #      [0, 0, 0, 0, 1],
    #      [0, 0, 0, 0, 0]]
    a = np.asarray(a)
    b = (a + a.T) / 2
    d, q = np.linalg.eig(b)
    d[d < 0] = 0
    x = q.dot(d.reshape((d.size, 1)) * q.T)
    return (x + x.T) / 2


def nearestPD(A):
    """Find the nearest positive-definite matrix to input

    A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code [1], which
    credits [2].

    [1] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd

    [2] N.J. Higham, "Computing a nearest symmetric positive semidefinite
    matrix" (1988): https://doi.org/10.1016/0024-3795(88)90223-6
    """
    B = (A + A.T) / 2
    _, s, V = np.linalg.svd(B)
    H = np.dot(V.T, np.dot(np.diag(s), V))
    A2 = (B + H) / 2
    A3 = (A2 + A2.T) / 2
    if isPD(A3):
        return A3
    spacing = np.spacing(np.linalg.norm(A))
    # The above is different from [1]. It appears that MATLAB's `chol` Cholesky
    # decomposition will accept matrixes with exactly 0-eigenvalue, whereas
    # Numpy's will not. So where [1] uses `eps(mineig)` (where `eps` is Matlab
    # for `np.spacing`), we use the above definition. CAVEAT: our `spacing`
    # will be much larger than [1]'s `eps(mineig)`, since `mineig` is usually on
    # the order of 1e-16, and `eps(1e-16)` is on the order of 1e-34, whereas
    # `spacing` will, for Gaussian random matrixes of small dimension, be on
    # othe order of 1e-16. In practice, both ways converge, as the unit test
    # below suggests.
    Imatrix = np.eye(A.shape[0])
    k = 1
    while not isPD(A3):
        mineig = np.min(np.real(np.linalg.eigvals(A3)))
        A3 += Imatrix * (-mineig * k**2 + spacing)
        k += 1
    return A3


def isPD(B):
    """Returns true when input is positive-definite, via Cholesky"""
    try:
        _ = np.linalg.cholesky(B)
        return True
    except np.linalg.LinAlgError:
        return False
