import numpy as np
from scipy.signal import find_peaks


def min_cell_cycle(dt):
    # minimal cell cycle of 50 minutes
    return int(50 / dt)


###


def add_cyto_pts(dl, column_name="cyto_pts", min_time_from_budding=30):
    """
    Predicts points of cytokinesis, which are added to the dataframe.

    Arguments
    --------
    dl: instance of dataloader
    column_name: string (optional)
        Name of column to store the predicted points of cytokinesis
    min_time_from_budding : float (optional)
        Time in minutes from budding, both before and after, where a point of
        cytokinesis cannot be predicted
    """
    dt = dl.dt
    dl.df[column_name] = np.nan
    # loop through lineages
    for idx in dl.ids:
        (t, bgr, bgrvar, mgr, mgrvar, buddings,) = dl.get_lineage_data(
            idx,
            [
                "time",
                "bud_growth_rate",
                "bud_growth_rate_var",
                "mother_growth_rate",
                "mother_growth_rate_var",
                "buddings",
            ],
        )
        cyto_pts = np.zeros(t.size)
        # find one cyto_point between a pair of buddings
        bs = np.where(buddings)[0]
        for start, end in zip(bs, bs[1:]):
            if end > start + min_cell_cycle(dt):
                # predict cyto_pt
                pred = find_cyto_pts_between_buddings(
                    bgr[start:end],
                    bgrvar[start:end],
                    mgr[start:end],
                    mgrvar[start:end],
                    offs=int(min_time_from_budding / dt),
                )
                if not np.isnan(pred):
                    cyto_pts[start + pred] = 1
        dl.df.loc[dl.df.id == idx, [column_name]] = cyto_pts


def add_G1(dl):
    """
    Indicates where the mother cells in likely to be in G1 using the
    predicted buddings and points of cytokinesis.
    Adds a boolean "G1" column to the dataframe.

    Arguments
    ---------
    dl: instance of dataloader
    """
    if "cyto_pts" not in dl.df.columns:
        print("No points of cytokinesis have been predicted.")
        print("Run add_cyto_pts")
    else:
        dl.df["G1"] = np.nan
        for idx in dl.ids:
            t, buddings, cyto_pts = dl.get_lineage_data(
                idx, ["time", "buddings", "cyto_pts"]
            )
            G1 = np.zeros(t.size)
            bs = np.where(buddings)[0]
            for start, end in zip(bs, bs[1:]):
                cyto = start + np.where(cyto_pts[start:end])[0]
                if np.any(cyto):
                    G1[cyto[0] : end] = 1
            dl.df.loc[dl.df.id == idx, ["G1"]] = G1


###


def find_cyto_pts_between_buddings(
    bgr,
    bgrvar,
    mgr,
    mgrvar,
    offs,
    thresh=100,
):
    """
    Find cytokinesis between buddings using the growth rates of mother and bud.
    A point of cytokinesis has both

        np.sqrt(bgrvar + mgrvar + (mgr - bgr)**2) / np.nanmean(mgr) <= thresh

    and

        bgr / np.nanmean(mgr)

    at a minimum.

    where thresh is incrementally increased until a point of cytokinesis is
    found.

    Parameters
    ----------
    bgr: array
        Bud's growth rate
    bgrvar : array
        Variance in the bud's growth rate
    mgr : array
        Mother's growth rate
    mgrvar: array
        Variance in the mother's growth rate
    offs: integer
        Offset (number of data points) for which data is ignored at the
        beginning and end of the time series so that the predicted point of
        cytokinesis cannot have either of these values.
    max_thresh: float (optional)
        Maximal value of the threshold

    Returns
    -------
    cyto_pt: integer or NaN
        The index of the predicted point of cytokinesis.
        NaN is returned if no point is found.
    """
    s1 = np.sqrt(bgrvar + mgrvar + (mgr - bgr) ** 2) / np.nanmean(mgr)
    s2 = bgr / np.nanmean(mgr)
    # prediction cannot be too close to a budding event
    s1 = s1[offs:-offs]
    s2 = s2[offs:-offs]
    # find cyto point
    res = np.where(s1 <= thresh)[0]
    if len(res) > 0:
        # return res[np.argmin(s1[res])]
        cyto = res[np.argmin(s2[res])]
    else:
        cyto = np.nan
    cyto_pt = cyto + offs
    return cyto_pt


###


def add_cyto_pts_from_Myo1(dl, signal, column_name, myo1_range_thresh=0.1):
    """
    Finds peaks in Myo1 signal stored in a dataloader instance.
    Peaks are added to dl.df as a column called "cyto_pts".

    Arguments
    ---------
    dl: dataloader instance
    signal: string
        Signal to use for Myo1, such as "max5px_med" or "nuc_est_conv"
    column_name : string
        Name of column where peaks in Myo1 will be stored.
    myo1_range_thresh: float (optional)
        Only lineages with a range in the Myo1 signal larger than this value
        will be considered.
    """
    dt = dl.dt
    dl.df[column_name] = np.nan
    for idx in dl.ids:
        myo1, t = dl.get_lineage_data(idx, [signal, "time"])
        cyto_points = np.zeros(t.size)
        myo1_range = np.nanquantile(myo1, 0.75) - np.nanquantile(myo1, 0.25)
        # only use data with a substantial response from Myo1
        if myo1_range > myo1_range_thresh:
            pks, pk_props = find_peaks(
                myo1,
                prominence=0.5 * myo1_range,
                distance=min_cell_cycle(dt),
            )
            cyto_points[pks] = 1
        dl.df.loc[dl.df.id == idx, [column_name]] = cyto_points
