import numpy as np


def by_thresh(grdiff, bgr, start, thresh, max_thresh):
    """
    Recursively define a crossing point based on a threshold range.

    Requirements for a valid 'crossing' point:
    * the difference between mother and bud growth rate (grdiff) is less
        than the threshold
    * the bud growth rate (bgr) is greater than the threshold

    The range of thresholds from thresh to max_thresh is checked from low to
    high.
    As soon as a point satisfies the requirements above, a value is returned.
    If there is no valid point for any of the thresholds within the range,
    NaN is returned.
    """
    zero_crossings = np.where((np.abs(grdiff) <= thresh) & (bgr >= thresh))[0]
    if len(zero_crossings) > 0:
        return zero_crossings[0] + start
    elif thresh < max_thresh:
        return by_thresh(grdiff, bgr, start, thresh + 1, max_thresh)
    else:
        return np.nan


def growth_rate_crossing_mean(d_gr, m_gr, start, end, ofs=2, lb=2, ub=15):
    """
    Find cytokinesis by growth rate crossing within a range of thresholds.

    This method enforces that cytokinesis is not within ofs time points of
    start or end.
    It only uses the mean growth rate values.
    """
    ostart = start + ofs
    oend = end - ofs
    diff = d_gr[ostart:oend] - m_gr[ostart:oend]
    cyto = by_thresh(diff, d_gr[ostart:oend], ostart, lb, ub)
    return cyto


def gaussiandist(x, var_x, y, var_y):
    """
    Distance between two gaussian variables given mean and standard
    deviation
    """
    return np.sqrt(var_x + var_y + (x - y) ** 2)


def growth_rate_crossing(
    bgr,
    bvar,
    mgr,
    mvar,
    start,
    end,
    ofs=2,
    thresh=2,
    max_thresh=15,
):
    """
    Find cytokinesis by growth rate crossing within a range of thresholds,
    noisy.

    This method enforces that cytokinesis is not within ofs time points of
    start or end.
    It uses both the mean and standard deviation of growth rate.
    """
    ostart = start + ofs
    oend = end - ofs
    grdiff = gaussiandist(
        bgr[ostart:oend],
        bvar[ostart:oend],
        mgr[ostart:oend],
        mvar[ostart:oend],
    )
    cyto = by_thresh(grdiff, bgr[ostart:oend], ostart, thresh, max_thresh)
    return cyto
