import numpy as np
import pandas as pd
from collections import namedtuple


###


def toarray(y):
    if type(y) == pd.core.frame.DataFrame:
        return y.to_numpy()
    else:
        return y


def todf(y, exampledf):
    if type(y) == pd.core.frame.DataFrame or exampledf is None:
        return y
    else:
        return pd.DataFrame(
            y, index=exampledf.index, columns=exampledf.columns
        )


###


def lbootstrap(data, statistic, noresamples=100, *args, **kwargs):
    """
    Uses statistical bootstrapping to estimate errors in a statistic.

    Parameters
    ----------
    data:   array
        A design matrix of data, with features as columns and replicates
        as rows
    statistic:  function returning a float
        Returns the desired statistic
    noresamples: integer
        The number of bootstrapped samples generated (by resampling with
        replacement)
    args, kwargs:
        Any additional arguments required by the statistic function

    Returns
    -------
    res:    a named tuple with fields
        median: float
            The median of the statistic calculated from the resampled data sets
        mean: float
            The mean of the statistic
        std: float
            The standard deviation of the statistic
        cf: list of two floats
            The 95% confidence interval
        iqr: float
            The interquartile range
    """
    if data.ndim == 1:
        data = data.reshape(data.size, -1)
    nosamples = data.shape[0]
    rsams = np.random.choice(nosamples, size=(nosamples, noresamples))
    stats = [
        statistic(data[rsams[:, i], :], *args, **kwargs)
        for i in np.arange(noresamples)
    ]
    Res = namedtuple("Res", ["cf", "std", "iqr", "median", "mean", "orig"])
    return Res(
        [np.nanquantile(stats, 0.025), np.nanquantile(stats, 0.975)],
        np.nanstd(stats),
        np.nanquantile(stats, 0.75) - np.nanquantile(stats, 0.25),
        np.nanmedian(stats),
        np.nanmean(stats),
        statistic(data),
    )


###


def entropy_time_series(data):
    """
    Estimates the noise in a collection of time series by calculating
    the entropy assuming that the data obey a multivariate normal
    distribution.

    Uses the Ledoit-Wolf estimator of the covariance if the empirical
    covariance is not positive definite.

    Arguments
    ---------
    data: array
        An array of time series with replicates as rows and time points
        as columns.

    Returns
    -------
    gaussianentropy: float
        An estimate of the entropy.
    cov: array
        An estimate of the covariance matrix.
    """
    # number of time points
    n = data.shape[1]
    # deviation from mean over replicates at each time point for each replicate
    ddev = data - np.nanmean(data, axis=0).reshape((1, n))
    # find covariance matrix
    cov = np.zeros((n, n))
    for i in np.arange(n):
        for j in np.arange(i + 1, n):
            cov[i, j] = np.nanmean(ddev[:, i] * ddev[:, j])
    cov = cov + cov.T + np.diag(np.nanvar(data, axis=0))
    # find determinant using cholesky decomposition
    if np.any(np.linalg.eigvals(cov) < 0):
        print("Using Ledoit-Wolf estimator")
        from sklearn.covariance import LedoitWolf

        cleandata = data[~np.isnan(data).any(axis=1), :]
        print(
            " dropping",
            data.shape[0] - cleandata.shape[0],
            "time series with NaNs",
        )
        print(" keeping", cleandata.shape[0], "time series")
        cov = LedoitWolf().fit(cleandata).covariance_
    try:
        ell = np.linalg.cholesky(2 * np.pi * np.e * cov)
        gaussianentropy = np.sum(np.log(np.diag(ell)))
        return gaussianentropy, cov
    except np.linalg.LinAlgError:
        print("Error: the covariance matrix is not positive definite")
        return np.nan, np.nan
