import matplotlib.pylab as plt
import numpy as np
from scipy.interpolate import interp1d

min_cell_cycle_in_tps = 3


def drop_label(df, label="mother_label"):
    """Drop label from the index of data frame."""
    if df is not None and label in df.index.names:
        df = df.reset_index(label).drop(columns=[label])
    return df


def correct_buds(buddings, bud_data, figs, bud_interpolate_indices=None):
    """
    Correct buddings and bud data.

    Drop adjacent buddings and interpolate over NaNs for specified signals.

    Parameters
    ----------
    buddings: pd.DataFrame
        Data frame of time series of budding events.
    bud_data: list of pd.DataFrame
        List of data frames of time series of bud signals.
        Bud volume should be the first element of the list.
    figs: boolean
        If True, plot volumes of all buds marking budding events with
        a red circle.
    bud_interpolate_indices: list of int
        Indices for signals in bud_data where linear interpolation should
        be used to remove NaNs within the time series for each bud.
    """
    # remove mother label
    bud_data = [drop_label(bud_dataset) for bud_dataset in bud_data]
    # check buddings and bud_data
    new_buddings, new_bud_data = skip_buddings(buddings, bud_data)
    # interpolate
    if bud_interpolate_indices is not None:
        for i in bud_interpolate_indices:
            if new_bud_data[i] is not None:
                for ind in new_buddings.index:
                    success, new_bud_series = interp_bud_series(
                        ind,
                        new_buddings,
                        new_bud_data[i],
                        figs,
                    )
                    if success:
                        new_bud_data[i].loc[ind] = new_bud_series
    return new_buddings, new_bud_data


def skip_buddings(buddings, bud_data):
    """
    Skip budding events too close to others.

    Two adjacent buddings events are implausible.
    We ignore the one corresponding to the larger bud volume.

    Parameters
    ----------
    buddings: pd.DataFrame
        Data frame of time series of budding events.
    bud_data: list of pd.DataFrame
        List of data frames of time series of bud signals.
        Bud volume should be the first element of the list.
    """
    bud_volume = bud_data[0]
    new_buddings = buddings.copy()
    new_bud_data = [
        bud_dataset.copy() if bud_dataset is not None else None
        for bud_dataset in bud_data
    ]
    for ind in buddings.index:
        ibuddings = np.concatenate(
            (np.where(buddings.loc[ind] > 0)[0], [buddings.size])
        )
        for ib, ib_end in zip(ibuddings, ibuddings[1:]):
            if ib_end < ib + min_cell_cycle_in_tps:
                print(" Skipping budding event for", ind)
                try:
                    vol = bud_volume.loc[ind].values.astype("float")
                except KeyError:
                    continue
                # ignore budding event with larger volume
                if vol[ib] > vol[ib_end]:
                    # ignore earlier budding
                    new_buddings.loc[ind].iloc[ib] = 0
                    for j in range(len(new_bud_data)):
                        if new_bud_data[j] is not None:
                            new_bud_data[j].loc[ind].iloc[ib] = np.nan
                else:
                    # ignore later budding
                    new_buddings.loc[ind].iloc[ib_end] = 0
    print()
    return new_buddings, new_bud_data


def interp_bud_series(ind, buddings, bud_signal, figs):
    """Use linear interpolation to fill NaN in time series of bud signals."""
    ibuddings = np.where(buddings.loc[ind] > 0)[0]
    ibuddings_end = np.concatenate((ibuddings, [buddings.size]))
    success = False
    if np.any(ibuddings):
        try:
            bs = bud_signal.loc[ind]
        except KeyError:
            print(f" No bud signal for {str(ind)}")
            return success, np.nan
        t = bs.index.values.astype("float")
        sig = bs.values.astype("float")
        # interpolate to fix
        new_bs = bs.copy()
        for ib, ib_end in zip(ibuddings_end, ibuddings_end[1:]):
            # check individual buds
            if ib_end > ib + min_cell_cycle_in_tps:
                # interpolate to fix any NaN
                bud_sig = sig[ib:ib_end]
                bud_t = t[ib:ib_end]
                # find last non-NaN data point
                ib_real = np.max(np.where(~np.isnan(bud_sig))[0])
                t_ib = bud_t[: ib_real + 1][~np.isnan(bud_sig[: ib_real + 1])]
                s_ib = bud_sig[: ib_real + 1][
                    ~np.isnan(bud_sig[: ib_real + 1])
                ]
                int_bud_sig = interp1d(t_ib, s_ib)
                # interpolate missing NaN within bud time series
                new_bs[ib : ib + ib_real + 1] = int_bud_sig(
                    t[ib : ib + ib_real + 1]
                )
        # plot
        if figs:
            plot_buds(ind, buddings, bud_signal, new_bs)
        # return interpolated values
        success = True
        return success, new_bs
    else:
        # no buddings
        return success, np.nan


def plot_buds(ind, buddings, bud_signal, new_bs=False):
    """
    Plot a given index from a bud signal data frame showing the
    signal and budding events for all buds of a mother.

    Boolean new_bs shows new interpolated bud signals.
    """
    ibuddings = np.where(buddings.loc[ind] > 0)[0]
    bs = bud_signal.loc[ind]
    plt.figure()
    for ib in ibuddings:
        # budding events
        plt.plot(bs.index[ib], bs.values[ib], "ro")
    ibs = np.concatenate(([0], ibuddings, [buddings.size]))
    for j, (start, end) in enumerate(zip(ibs, ibs[1:])):
        if np.any(new_bs):
            # plot interpolated data differently
            plt.plot(new_bs[start:end], "b--", alpha=0.5)
        plt.plot(bs[start:end], "b.-")
    plt.xlabel("time")
    plt.ylabel("bud signal")
    plt.grid()
    plt.title(str(ind))
    plt.show(block=False)
