import os
import pprint
from pathlib import Path

import numpy as np
import pandas as pd

from wela.correct_buds import correct_buds

try:
    from postprocessor.grouper import Grouper

    from wela.add_bud_fluorescence import add_bud_fluorescence
except ModuleNotFoundError:
    print("Can only load tsv files - cannot find aliby's postprocessor.")


class dataloader:
    """
    Compile aliby data sets into a single extensible data frame.

    Buddings and bud_volume are always loaded first.

    Parameters
    ----------
    h5dir: string
        The directory where data are stored as h5 files with one
        directory per experiment
    wdir: string
        The working directory where .tsv files will be saved
    ls: boolean
        If True, list files and directories available.

    Example
    -------
    >>> from dataloader import dataloader
    >>> dl = dataloader(h5dir="experiments", wdir=".")
    >>> dl.load(key_index = "median_GFP", cutoff=0.8)

    to select cells that have non-NaN values for median_GFP for 80%
    of the experiment.

    The names of experiments and data sets are stored in dictionaries
    to avoid cutting and pasting of names.

    To load an experiment from the data directory, first use,
    for example,

    >>> dl.load(dl.experiments[0])

    or give the name of the directory explicitly

    >>> dl.load("experiment_directory")

    which will load the data from h5 files.

    Once the loading has been complete, the data is saved in a
    text file - a .tsv file - in the working directory.

    To load this file in future use, for example

    >>> dl.load(dl.datasets[0])

    or give the name of the file explicitly

    >>> dl.load("data_set.tsv")

    To see the data, use

    >>> dl.df

    which can be plotted directly with seaborn.

    >>> import seaborn as sns
    >>> sns.relplot(x="time", y="median_GFP", hue="group",
            kind="line", data=dl.df)
    """

    def __init__(self, h5dir=None, wdir=".", ls=True):
        # from grouper.signals to abbreviations
        self.g2a_dict = self.define_g2a_dict()
        # from abbreviations to grouper.siglist
        self.a2g_dict = {v: k for (k, v) in self.g2a_dict.items()}
        # establish paths
        self.wdirpath = Path(wdir)
        if h5dir is None:
            self.h5dirpath = None
        else:
            self.h5dirpath = Path(h5dir)
        if ls:
            self.ls

    def define_g2a_dict(self):
        """Dict to convert names of grouper.signals to short forms."""
        g2a_dict = {
            "extraction/cy5/max/median": "cy5",
            "extraction/general/None/area": "area",
            "extraction/general/None/volume": "volume",
            "extraction/general/None/image_x": "image_x",
            "extraction/general/None/image_y": "image_y",
            "extraction/general/None/absolute_x": "absolute_x",
            "extraction/general/None/absolute_y": "absolute_y",
            "extraction/Flavin_bgsub/max/mean": "flavin",
            "postprocessing/buddings/extraction_general_None_volume": "buddings",
            "postprocessing/bud_metric/extraction_general_None_volume": "bud_volume",
        }
        for fl in ["GFP", "mCherry"]:
            g2a_dict[f"extraction/{fl}/max/median"] = f"median_{fl}"
            g2a_dict[f"extraction/{fl}/max/mean"] = f"mean_{fl}"
            g2a_dict[f"extraction/{fl}/max/std"] = f"std_{fl}"
            g2a_dict[f"extraction/{fl}/max/max5px_median"] = (
                f"max5px_median_{fl}"
            )
            g2a_dict[f"extraction/{fl}/max/nuc_est_conv"] = (
                f"nuc_est_conv_{fl}"
            )
            g2a_dict[
                f"postprocessing/bud_metric/extraction_{fl}_max_median"
            ] = f"bud_median_{fl}"
            g2a_dict[f"postprocessing/bud_metric/extraction_{fl}_max_mean"] = (
                f"bud_mean_{fl}"
            )
            g2a_dict[f"postprocessing/bud_metric/extraction_{fl}_max_std"] = (
                f"bud_std_{fl}"
            )
        return g2a_dict

    @property
    def ls(self):
        """
        List the experiments in h5dir and the datasets in wdir.

        The names of all directories in h5dir are stored in the .experiments
        dictionary.

        The names of all datasets in wdir are stored in the .datasets
        dictionary.
        """
        pp = pprint.PrettyPrinter()
        if self.h5dirpath is not None:
            # find h5 data
            print("\nData directory is", str(self.h5dirpath.resolve()))
            print("Experiments available:")
            dirs = [f.name for f in self.h5dirpath.glob("*") if f.is_dir()]
            # directories of data are stored in experiments
            self.experiments = {i: name for i, name in enumerate(dirs)}
            pp.pprint(self.experiments)
        # find processed data
        print("\nWorking directory is", str(self.wdirpath.resolve()))
        print("Datasets available:")
        files = [f.name for f in self.wdirpath.glob("*.tsv") if f.is_file()]
        # tsv files are stored in datasets
        self.datasets = {i: name for i, name in enumerate(files)}
        pp.pprint(self.datasets)

    def get_grouper(self, dataname):
        """
        Return an instance of Grouper.

        Arguments
        ---------
        dataname: string
            The name of the dataset

        Returns
        -------
        grouper: instance of Grouper
        """
        if dataname[-4:] == ".tsv":
            dataname = dataname[:-4]
        grouper = Grouper(self.h5dirpath / dataname)
        return grouper

    def update_dictionaries(self, extra_g2a_dict, overwrite_dict):
        """Update conversion dictionaries."""
        if extra_g2a_dict and not overwrite_dict:
            self.g2a_dict = {**self.g2a_dict, **extra_g2a_dict}
        elif extra_g2a_dict:
            self.g2a_dict = extra_g2a_dict
        for key, value in zip(
            [
                "postprocessing/buddings/extraction_general_None_volume",
                "postprocessing/bud_metric/extraction_general_None_volume",
            ],
            ["buddings", "bud_volume"],
        ):
            if key not in self.g2a_dict:
                self.g2a_dict[key] = value
        self.a2g_dict = {v: k for (k, v) in self.g2a_dict.items()}

    def include_bud_fluorescence(self, grouper, dataname):
        """Add mean and median bud fluorescence to the h5 files."""
        # find fluorescence channels
        channels = list(grouper.channels)
        channels.remove("Brightfield")
        signals = [
            signal
            for two_signal in [
                [
                    f"/extraction/{channel}/max/median",
                    f"/extraction/{channel}/max/mean",
                ]
                for channel in channels
            ]
            for signal in two_signal
        ]
        # add bud fluorescence to h5 files
        add_bud_fluorescence(self.h5dirpath / dataname, signals)

    def load(
        self,
        dataname,
        key_index="buddings",
        cutoff=0.8,
        interpolate_list=None,
        extra_g2a_dict=None,
        pxsize=0.182,
        use_tsv=False,
        overwrite_dict=False,
        hours=True,
        bud_fluorescence=False,
        tmax_in_mins_dict=None,
    ):
        """
        Load either an experiment from h5 files or a tsv data set.

        Data are stored in a long data frame with the .df attribute.
        The 'time' variable becomes a column.

        New data is added to the existing data frame.

        Parameters
        ----------
        dataname: string
            Either the name of the directory with h5 files or a
            file name for a tsv data set.
        key_index: string
            Short name for key record that will be used to select cells.
        cutoff: float
            Select cells for key record that remain for at least cutoff
            fraction of the experiment's total duration.
            A cutoff of zero means select all cells.
        interpolate_list: list of str
            A list of signals for which linear interpolation will replace
            NaNs
        extra_g2a_dict: dictionary (optional)
            A dictionary of extra parameters to extract, which relates the
            aliby name to an abbreviation
        pxsize: float
            Pixel size in microns, which is used to convert volumes.
        use_tsv: boolean
            If True, always load the data from a tsv file.
        overwrite_dict: boolean
            If True, overwrite the internal dictionary with extra_g2a_dict.
        hours: boolean
            If True, convert times to hours (dividing by 60).
        bud_fluorescence: boolean
            If True, add mean and median bud fluorescence to the data frame.
        tmax_in_mins_dict: dict (optional)
            A dictionary with positions as keys and maximum times in minutes as
            values. For example: { "PDR5_GFP_001": 6 * 60}.
            Data will only be include up to this time point, a way to
            avoid errors in assigning lineages because of clogging.

        Returns
        -------
        grouper: aliby grouper object
            grouper for an experiment if one is loaded

        Example
        -------
        >>> dl.load(dataname, {"extraction/cy5/max/median" : "cy5"})
        >>> dl.load(dataname, key_index="flavin", cutoff= 0.9)
        """
        if not len(dataname):
            raise Exception("No file to load.")
        elif dataname[-4:] == ".tsv":
            dataname = dataname[:-4]
            use_tsv = True
        # root name
        self.dataname = dataname
        if use_tsv:
            self.load_tsv(dataname)
        else:
            # update conversion dictionaries
            self.update_dictionaries(extra_g2a_dict, overwrite_dict)
            # create instance of grouper
            grouper = self.get_grouper(dataname)
            # find time interval between images
            self.dt = grouper.tinterval
            print("\n---\n" + dataname + "\n---")
            if bud_fluorescence:
                # call postprocessor to add bud fluorescence to h5 files
                self.include_bud_fluorescence(grouper, dataname)
            print("Signals available:")
            for signal in grouper.available:
                print(" ", signal)
            print()
            if self.a2g_dict[key_index] not in grouper.available:
                raise Exception(f"The key index {key_index} is unavailable.")
            # get indices for all buds
            bud_indices = self.get_bud_indices(grouper, key_index)
            # get key_index data
            r_df = self.get_key_index_data(
                grouper=grouper,
                key_index=key_index,
                bud_indices=bud_indices,
                cutoff=cutoff,
                tmax_in_mins_dict=tmax_in_mins_dict,
            )
            # add data for other signals to data for key_index
            r_df = self.load_h5(
                grouper=grouper,
                key_index=key_index,
                bud_indices=bud_indices,
                r_df=r_df,
                interpolate_list=interpolate_list,
                tmax_in_mins_dict=tmax_in_mins_dict,
            )
            if pxsize:
                # convert volumes to micron^3
                for signal in r_df.columns:
                    if "volume" in signal:
                        r_df[signal] *= pxsize**3
            if hours:
                r_df.time /= 60
            # create new attribute or merge with existing one
            if hasattr(self, "df"):
                self.df = pd.merge(self.df, r_df, how="left")
            else:
                self.df = r_df
        print(f"Data size is {self.df.shape}.")
        # define ids
        self.ids = list(self.df.id.unique())
        if not use_tsv:
            return grouper

    def load_h5(
        self,
        grouper,
        key_index,
        bud_indices,
        r_df,
        interpolate_list,
        tmax_in_mins_dict,
    ):
        """Load data from h5 files into one long data frame."""
        # load other signals
        print("\nGetting data for other signals...")
        for i, sigpath in enumerate(self.g2a_dict):
            if (
                sigpath in grouper.available
                and not ("buddings" in sigpath or "bud_metric" in sigpath)
                and sigpath != self.a2g_dict[key_index]
            ):
                record = grouper.concat_signal(
                    sigpath,
                    cutoff=0,
                    tmax_in_mins_dict=tmax_in_mins_dict,
                )
                record = self.remove_buds(record, bud_indices)
                # interpolate to remove internal NaNs for signals from mothers
                if (
                    interpolate_list
                    and self.g2a_dict[sigpath] in interpolate_list
                ):
                    record = interp_signal(record)
                # convert to long data frame
                tdf = self.long_df_with_id(record, self.g2a_dict[sigpath])
                # merge into one data set
                # cell IDs not in r_df will be automatically dropped
                r_df = pd.merge(r_df, tdf, how="left")
        print("Getting bud data...")
        # load and correct buddings and bud_volume
        b_df = self.load_bud_data(
            grouper=grouper,
            figs=False,
            interpolate_list=interpolate_list,
            tmax_in_mins_dict=tmax_in_mins_dict,
        )
        r_df = pd.merge(r_df, b_df, how="left")
        return r_df

    def get_key_index_data(
        self, grouper, key_index, bud_indices, cutoff, tmax_in_mins_dict
    ):
        """
        Find data for the key record.

        Cells must be retained at least a cutoff fraction of the
        experiment's duration.
        """
        key_index_path = self.a2g_dict[key_index]
        print("Getting data for key index...")
        record = grouper.concat_signal(
            key_index_path, cutoff=cutoff, tmax_in_mins_dict=tmax_in_mins_dict
        )
        record = self.remove_buds(record, bud_indices)
        if record is not None:
            r_df = self.long_df_with_id(record, key_index)
            return r_df
        else:
            raise Exception(f"{key_index_path} cannot be found.")

    def get_bud_indices(self, grouper, key_index):
        """Use key_index to get a multi-index for all buds."""
        key_index_path = self.a2g_dict[key_index]
        print("Getting data to identify bud indices...")
        record = grouper.concat_signal(key_index_path, mode="raw_daughters")
        bud_indices = record.droplevel("mother_label").index
        return bud_indices

    def remove_buds(self, df, bud_indices):
        """Remove rows that are buds from a data frame."""
        new_index = df.index.difference(bud_indices)
        return df.loc[new_index]

    def load_bud_data(
        self,
        grouper,
        figs,
        interpolate_list,
        tmax_in_mins_dict,
    ):
        """
        Load buddings, bud volume, and any other bud signals.

        Bud signals are indexed by their mother cells.

        Drop adjacent buddings and interpolate specified bus signals
        to remove NaNs.
        """
        bud_signals = [
            key for key in list(self.a2g_dict.keys()) if "bud_" in key
        ]
        # bud_volume must be the first signal
        bud_signals.insert(0, bud_signals.pop(bud_signals.index("bud_volume")))
        # find indices of bud signals to be interpolated
        if interpolate_list:
            bud_interpolate_indices = [
                i
                for i in range(len(bud_signals))
                if bud_signals[i] in interpolate_list
            ]
        else:
            bud_interpolate_indices = None
        # load buddings
        buddings = grouper.concat_signal(
            self.a2g_dict["buddings"],
            cutoff=0,
            tmax_in_mins_dict=tmax_in_mins_dict,
        )
        # bud_volume and any other signals; missing signals return None
        bud_data = [
            grouper.concat_signal(
                self.a2g_dict[bud_signal],
                cutoff=0,
                tmax_in_mins_dict=tmax_in_mins_dict,
            )
            for bud_signal in bud_signals
        ]
        # perform correction
        new_buddings, new_bud_data = correct_buds(
            buddings, bud_data, figs, bud_interpolate_indices
        )
        # convert to long df and merge
        df = self.long_df_with_id(new_buddings, "buddings")
        # merging will preserve data for cells in buddings
        for new_bud_dataset, bud_signal in zip(new_bud_data, bud_signals):
            if new_bud_dataset is not None:
                df = pd.merge(
                    df,
                    self.long_df_with_id(new_bud_dataset, bud_signal),
                    how="left",
                )
        return df

    def load_tsv(self, dataname):
        """
        Load data from a tsv file.

        Parameters
        ----------
        dataname: string
            Name of the file
        """
        r_df = pd.read_csv(str(self.wdirpath / (dataname + ".tsv")), sep="\t")
        print("\nLoaded", dataname)
        if hasattr(self, "r"):
            # merge new data to current dataframe
            self.df = pd.merge(self.df, r_df, how="left")
        else:
            self.df = r_df
        # store time interval
        self.dt = np.min(np.diff(np.sort(self.df.time.unique())))

    def save(self, dataname=None):
        """
        Save the .df data frame to a tab separated value (tsv) file.

        Parameters
        ----------
        dataname: string (optional)
            Name of the file
        """
        if dataname is None:
            dataname = self.dataname
        self.df.to_csv(
            str(self.wdirpath / (dataname + ".tsv")), sep="\t", index=False
        )
        print("Saved", dataname)

    def long_df_with_id(self, df, char_name):
        """Convert an aliby multi-index data frame into a long data frame."""
        df = long_df(df, char_name)
        # add unique id for each cell
        if (
            "position" in df.columns
            and "cell_label" in df.columns
            and "trap" in df.columns
        ):
            df.insert(
                0,
                "id",
                df["position"]
                + ";"
                + df["cell_label"].astype(str)
                + ";"
                + df["trap"].astype(str),
            )
        # drop redundant columns
        df = df.drop(["position", "trap", "cell_label"], axis=1)
        return df

    @property
    def revert_df(self):
        """
        Convert the 'id' column into aliby's standard columns and vice versa.

        These columns are 'position', 'trap', and 'cell_label', either
        adding three columns to the .df data frame or removing three
        columns.
        """
        if (
            "position" not in self.df.columns
            and "trap" not in self.df.columns
            and "cell_label" not in self.df.columns
        ):
            # add back dropped columns by expanding id column
            self.df[["position", "trap", "cell_label"]] = self.df.id.str.split(
                ";", expand=True
            )
        elif (
            "position" in self.df.columns
            and "trap" in self.df.columns
            and "cell_label" in self.df.columns
        ):
            # drop columns redundant with "id"
            self.df = self.df.drop(["position", "trap", "cell_label"], axis=1)
        else:
            print("Cannot revert.")

    def wide_df(self, signal, x="time", y="id"):
        """
        Pivot the .df data frame to return a standard aliby data frame.

        Parameters
        ----------
        signal: string
            The signal of interest
        x: string (default: 'time')
            The x value of interest
        y: string (default: 'id')
            The y value of interest

        Example
        -------
        >>> dl.wide_df("median_GFP")
        """
        return self.df.pivot(columns=x, index=y, values=signal)

    def sub_df(
        self,
        signal=None,
        dropna=False,
        duration_threshold=None,
        tmin=None,
        tmax=None,
    ):
        """
        Find a sub data frame of dataloader's main data frame.

        Parameters
        ----------
        signal: str
            The signal to which the duration will be applied.
        dropna: boolean
            If True, drop cells that have all NaN values for a given signal.
        duration_threshold: float
            Specifies the fraction of the total duration of the time-lapse
            experiment for which a cell must be present in a trap.
        tmin: float (optional)
            Only include data for times greater than tmin
        tmax: float (optional)
            Only include data for times less than tmax
        """
        sdf = self.df
        selected_ids = []
        # drop signals that are all NaN
        if dropna and signal:
            selected_ids = list(self.wide_df(signal).dropna(how="all").index)
        # apply duration threshold
        if duration_threshold and (
            duration_threshold > 0 and duration_threshold < 1
        ):
            if signal is None:
                print("A signal must be specified.")
            else:
                wdf = self.wide_df(signal)
                keep = (
                    wdf.notna().sum(axis=1)
                    > duration_threshold * wdf.columns.size
                ).values
                selected_ids = set(selected_ids).union(list(wdf.index[keep]))
        if selected_ids:
            sdf = self.df[self.df.id.isin(list(selected_ids))].copy()
        # restrict time
        if tmin and tmax:
            sdf = sdf[(sdf.time >= tmin) & (sdf.time <= tmax)]
        elif tmin:
            sdf = sdf[sdf.time >= tmin]
        elif tmax:
            sdf = sdf[sdf.time <= tmax]
        return sdf

    def get_time_series(self, signal, group=None, df=None):
        """
        Extract a signal as a 2D array with each row a time series.

        Parameters
        ----------
        signal: string
            The signal to be extracted

        Returns
        -------
        t: 1D array
            Time values
        values: 2D array
            The values of the signal with each row a separate time
            series
        group: string (optional)
            Selected data only for a particular group
        df: pd.DataFrame (optional)
            If specified, use this dataframe not dataloader's.
        """
        if df is None:
            if group is None:
                wdf = self.wide_df(signal, x="time", y="id")
            elif group in self.df.group.values:
                wdf = self.df[self.df.group == group].pivot(
                    columns="time", index="id", values=signal
                )
            else:
                print(group, "not recognised")
                return None, None
        else:
            if group is None:
                wdf = df.pivot(columns="time", index="id", values=signal)
            else:
                wdf = df[df.group == group].pivot(
                    columns="time", index="id", values=signal
                )
        time = wdf.columns.to_numpy()
        # each column is a time series
        values = wdf.to_numpy()
        return time, values

    def put_time_series(self, values, signal):
        """
        Insert a 2D array of data with each column a time series.

        Parameters
        ----------
        values: array
            The data to be inserted into the data frame.
        signal: string
            The name of the signal.
        """
        # find a suitable column in r to generate a pivot
        cols = list(self.df.columns)
        for col in ["id", "time"]:
            cols.remove(col)
        # use cols[0] to find index and columns of a pivot
        wdf = self.wide_df(cols[0])
        # create a dataframe using the structure of a pivot
        df = pd.DataFrame(values, index=wdf.index, columns=wdf.columns)
        # convert to long dataframe
        tdf = self.long_df_with_id(df, signal)
        # add to r
        self.df = pd.merge(self.df, tdf, how="left")

    @property
    def strains(self):
        """List all strains in df.group."""
        for strain in list(self.df.group.unique()):
            print(" " + strain)

    def get_lineage_data(self, idx, signals):
        """
        Return signals for a single lineage specified by idx.

        Parameters
        ----------
        idx: integer
            Element of self.ids.
        signals: list of strings
            Signals to be returned.
        """
        if isinstance(signals, str):
            signals = [signals]
        res = [
            self.df[self.df.id == idx][signal].to_numpy() for signal in signals
        ]
        return res


def long_df(df, char_name, var_name="time"):
    """Convert a multi-index dataframe into a long dataframe."""
    # flatten multi-index dataframe
    df = df.copy()
    if isinstance(df, pd.Series):
        df = df.unstack()
    hnames = df.index.names
    df.index = df.index.set_names(hnames)
    df.reset_index(inplace=True)
    # melt to create time column
    df = df.melt(id_vars=hnames, var_name=var_name, value_name=char_name)
    return df


def interp_signal(df):
    """Use interpolation to remove internal NaNs in data frame."""
    # wide aliby data frame
    df = df.interpolate("linear", axis=1, limit_area="inside")
    return df


def sort_df(df, signal, functionstr="sum"):
    """
    Return a sorted (categorical) data frame.

    Note the sorting will make the "id" column categorical, which is
    necessary for plotting via kymograph.

    Parameters
    ----------
    signal: string
        The signal to sort by.
    functionstr: string
        A function to apply to the signal before sorting.
        Default value is "sum" so that all the data for each cell
        is summed before sorting.
        Other options include "max", "mean", "median", etc.
    """
    if "id" not in df.columns:
        print("The data frame requires an id column.")
    else:
        sdf = df.groupby(["id"])
        # apply function to data grouped by ID
        if functionstr in dir(sdf):
            sdf = eval("sdf." + functionstr + "()")
        # sort
        sdf = sdf.sort_values(by=signal)
        sort_order = list(sdf.index)
        # sort original dataframe in the same way
        rdf = df.copy()
        rdf.id = pd.Categorical(rdf.id, categories=sort_order, ordered=True)
        rdf = rdf.sort_values(by="id")
        # return sorted dataframe
        return rdf


def find_full_omid(omid: str or int, outdir: str):
    """
    Find full name of experiment from an OMERO ID.

    Search outdir directory for name.
    """
    full_omid = [
        f
        for f in os.listdir(outdir)
        if os.path.isdir(os.path.join(outdir, f)) and str(omid) in f
    ].pop()
    return full_omid
