import numpy as np
import pandas as pd
from scipy import signal
from scipy.integrate import simpson


def classic_periodogram(data, sampling_interval, oversampling_factor=1):
    """
    Estimate the power spectral density using a periodogram.

    This is based on a fast Fourier transform.

    The power spectrum is normalised so that the area under the power
    spectrum is constant across different time series, thus allowing users
    to easily compare spectra across time series.  See:
    * Scargle (1982). Astrophysical Journal 263 p. 835-853
    * Glynn et al (2006). Bioinformatics 22(3) 310-316

    Parameters
    ---------
    data: 2D array
        Time series of measurement values as rows.
    sampling_interval: float
        Sampling interval of measurement values.
    oversampling_factor: float
        Indicates how many times a signal is sampled over the Nyquist rate.
        For example, if the oversampling factor is 2, the signal is sampled
        at twice the Nyquist rate.

    Returns
    -------
    freqs: 1D array
        Array of sample frequencies.
    power: 2D array
        Power spectral density or power spectrum of the time series.

    Example
    -------
    >>> rng = np.random.default_rng()
    >>> t = np.arange(0, 10, 0.1)
    >>> ys = np.stack([np.sin(2*np.pi*t) + 3*np.sin(4*np.pi*t) for _ in range(10)])
    >>> data = ys + rng.standard_normal(ys.shape)*0.3
    >>> f, p = classic_periodogram(data, 0.1)
    >>> plt.plot(f, p[3,:])
    """
    # interpolate any NaN between data values
    data = pd.DataFrame(data).interpolate(axis=1).to_numpy()
    # number of time points
    nt = data.shape[1]
    # parameters
    params = {
        "fs": 1 / (oversampling_factor * sampling_interval),
        "nfft": nt * oversampling_factor,
        "detrend": "constant",
        "return_onesided": True,
        "scaling": "spectrum",
    }
    # find periodogram
    if np.any(np.isnan(data)):
        power = np.empty((data.shape[0], int(nt / 2) + 1))
        for i in range(data.shape[0]):
            freqs, power[i, :] = signal.periodogram(
                data[i, ~np.isnan(data[i, :])], **params
            )
    else:
        freqs, power = signal.periodogram(data, **params)
    # required to correct unit
    freqs = oversampling_factor * freqs
    # normalisation (Scargle, 1982; Glynn et al., 2006)
    power = power * (0.5 * nt)
    # normalisation by variance to compare different time series (Scargle, 1982)
    power = power / np.nanvar(data, axis=1).reshape((data.shape[0], 1))
    return freqs, power


def snr(fft_freqs, fft_power, cutoff_freq):
    """
    Get signal-to-noise ratio from a Fourier spectrum.

    Get signal-to-noise ratio from a Fourier spectrum. Frequencies lower than
    the cut-off are considered signal; frequencies above the cut-off are
    considered noise. The signal-to-noise ratio is defined as the area under
    the Fourier spectrum to the left of the cut-off divided by the area under
    the Fourier spectrum to the right of the cut-off.

    Parameters
    ----------
    fft_freqs: 1D array
        The frequencies of the Fourier spectra.
    fft_power: 2D array
        The periodogram for time series as rows.
    cutoff_freq : float
        Cut-off frequency to divide signal from noise.
    """
    snr = np.empty(fft_power.shape[0])
    cutoff_index = np.argmin(np.abs(fft_freqs - cutoff_freq))
    for i in range(fft_power.shape[0]):
        area_all = simpson(y=fft_power[i, :], x=fft_freqs)
        area_signal = simpson(
            y=fft_power[i, 0:cutoff_index],
            x=fft_freqs[0:cutoff_index],
        )
        area_noise = area_all - area_signal
        snr[i] = area_signal / area_noise
    return snr


def lowpass_filter(data, cutoff, fs, N):
    """Apply low-pass filter to a single time-series."""
    data = data - np.mean(data)
    sos = signal.butter(N, cutoff, btype="lowpass", fs=fs, output="sos")
    filtered_data = signal.sosfiltfilt(sos, data)
    return filtered_data + np.mean(data)


def flavin_filter(data, cutoff=1 / 350, fs=1 / 5, N=2):
    """Apply high-pass Butterworth filter to time series arranged in rows."""
    # use a digital filter assuming regularly spaced sampling
    sos = signal.butter(N=N, Wn=cutoff, btype="highpass", fs=fs, output="sos")
    # fill NaNs between data points
    data = pd.DataFrame(data).interpolate(axis=1).to_numpy()
    # subtract time series by mean
    data_norm = data - np.nanmean(data, axis=1).reshape(data.shape[0], 1)
    if np.any(np.isnan(data_norm)):
        # filter one at a time because of NaNs at different times
        filtered_timeseries = np.nan * np.ones(data_norm.shape)
        for i in range(data_norm.shape[0]):
            s_data = data_norm[i, :]
            filtered = signal.sosfiltfilt(sos, s_data[~np.isnan(s_data)])
            filtered_timeseries[i, ~np.isnan(s_data)] = filtered
    else:
        # filter all at once
        filtered_timeseries = signal.sosfiltfilt(sos, data_norm, axis=1)
    return filtered_timeseries
