import pickle
from pathlib import Path

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import gaussianprocessderivatives as gp
from pathos.multiprocessing import ProcessingPool
import multiprocessing

from wela.dataloader import dataloader, long_df
from wela.figs2pdf import figs2pdf


def find_all_grs(
    dataname,
    max_no_cells=None,
    h5dir=None,
    wdir=".",
    use_tsv=True,
    bud_bounds=None,
    mother_bounds=None,
    multiprocess=False,
    convert_to_hours=True,
):
    """
    Find mother and bud growth rates.

    Parameters
    ----------
    dataname: str
        Name of experiment, passed to dataloader.
    max_no_cells: None or int
        The maximum number of cells to process.
    h5dir: str (optional)
        The directory containing the h5 files.
    wdir: str (optional)
        The directory containing any tsv files to load and where
        files will be saved.
    use_tsv: boolean (optional)
        If True, load initial data from a tsv file.
    bud_bounds: dict, optional
        To change bounds on the hyperparameters for the Gaussian process
        applied to buds.
    mother_bounds: dict, optional
        To change bounds on the hyperparameters for the Gaussian process
        applied to mothers.
    multiprocess: boolean, optional
        If True, use multiprocessing, but no plots are then produced.
    convert_to_hours: boolean
        If True, divide time values by 60.

    Example
    -------
    >>> from wela.growth_rate import find_all_grs
    >>> find_all_grs("Pdr5", max_no_cells=2, wdir="tsv_data")
    """
    wdirpath = Path(wdir)
    # only save if all cells are processed
    if max_no_cells is None or not max_no_cells:
        save = True
    else:
        save = False
    # get data
    dl = dataloader(h5dir=h5dir, wdir=wdir)
    dl.load(dataname, use_tsv=use_tsv)
    b_vol = dl.wide_df("bud_volume")
    m_vol = dl.wide_df("volume")
    buddings = dl.wide_df("buddings")
    # find bud growth rate
    default_bounds = {0: (0, 8)}
    if bud_bounds:
        bud_bounds = {**bud_bounds, **default_bounds}
    else:
        bud_bounds = default_bounds
    res_b = find_gr(
        b_vol,
        buddings,
        max_no_cells=max_no_cells,
        bounds=bud_bounds,
        multiprocess=multiprocess,
        convert_to_hours=convert_to_hours,
    )
    if save:
        with open(str(wdirpath / dataname) + "_res_b.pkl", "wb") as file:
            pickle.dump(res_b, file)
        figs2pdf(str(wdirpath / dataname) + "_b.pdf")
        plt.close("all")
    # find mother growth rate
    res_m = find_gr(
        m_vol,
        max_no_cells=max_no_cells,
        bounds=mother_bounds,
        multiprocess=multiprocess,
        convert_to_hours=convert_to_hours,
    )
    if save:
        with open(str(wdirpath / dataname) + "_res_m.pkl", "wb") as file:
            pickle.dump(res_m, file)
        figs2pdf(str(wdirpath / dataname) + "_m.pdf")
        plt.close("all")
    return res_b, res_m


def find_gr(
    volume_df,
    buddings_df=None,
    bounds=None,
    noruns=10,
    verbose=True,
    multiprocess=False,
    max_no_cells=None,
    convert_to_hours=True,
):
    """Use a Gaussian process to find single-cell growth rates."""
    # default bounds on hyperparameters
    if buddings_df is None:
        # mother params
        def_bounds = {0: (2, 10), 1: (2, 5), 2: (-4, 8)}
    else:
        # bud params
        def_bounds = {0: (2, 10), 1: (-1, 5), 2: (-4, 8)}
    if bounds:
        # merge with default bounds
        bounds = {**def_bounds, **bounds}
    else:
        # use default bounds
        bounds = def_bounds
    # get data from dataframe
    volume = volume_df.to_numpy()
    t = volume_df.columns.to_numpy()
    # convert to hours so that growth rate has the correct units
    if convert_to_hours:
        t /= 60
    # signals to be found
    if buddings_df is None:
        buddings = None
        signals = [
            "smoothed_volume",
            "growth_rate",
            "d_growth_rate",
        ]
    else:
        buddings = buddings_df.to_numpy()
        signals = [
            "smoothed_bud_volume",
            "bud_growth_rate",
            "d_bud_growth_rate",
        ]
    signals += [s + "_var" for s in signals]
    # store parameters for runGP
    config = {
        "bounds": bounds,
        "noruns": noruns,
        "signals": signals,
        "verbose": verbose,
    }
    if max_no_cells:
        no_cells = np.min([volume.shape[0], max_no_cells])
    else:
        no_cells = volume.shape[0]
    # loop over cells
    if multiprocess:
        config["verbose"] = False
        res = run_parallel(t, volume, buddings_df, buddings, no_cells, config)
    else:
        res = run_series(t, volume, buddings_df, buddings, no_cells, config)
    # convert results into dataframes
    res_df = {
        s: pd.DataFrame(
            res[s], index=volume_df.index, columns=volume_df.columns
        )
        for s in signals
    }
    return res_df


def run_series(t, volume, buddings_df, buddings, no_cells, config):
    """Run GP on each time series in turn."""
    signals = config["signals"]
    # store results in a dictionary
    res = {s: np.nan * np.ones(volume.shape) for s in signals}
    # loop over cells
    for idx in range(no_cells):
        ivol = volume[idx, :]
        if buddings_df is None:
            # data is a single time series
            ires = runGP(
                t,
                ivol,
                **config,
                title=f"mother time series {idx:d} / {no_cells-1:d}",
            )
            if ires:
                # store results as arrays
                for s in signals:
                    res[s][idx, :] = ires[s]
        else:
            # time series comprises bud trajectories
            bs = np.concatenate(([0], np.where(buddings[idx, :])[0], [t.size]))
            for j, (start, end) in enumerate(zip(bs, bs[1:])):
                # data for one bud
                x = t[start:end]
                y = ivol[start:end]
                # run GP
                ires = runGP(
                    x,
                    y,
                    **config,
                    title=f"bud time series {idx:d} / {no_cells-1:d}: bud {j:d} /"
                    f" {bs.size-2:d}",
                )
                if ires:
                    # store results as arrays
                    for s in signals:
                        res[s][idx, start:end] = ires[s]
    return res


def run_parallel(t, volume, buddings_df, buddings, no_cells, config):
    """
    Use multiprocessing to speed GPs.

    The CPU cores either run mothers or buds from the same mother.
    """
    batch_size = np.max([multiprocessing.cpu_count() - 1, 1])
    pool = ProcessingPool(nodes=batch_size)

    def batch_GP(x, y):
        res = runGP(x, y, **config)
        return res

    signals = config["signals"]
    # store results in a dictionary
    res = {s: np.nan * np.ones(volume.shape) for s in signals}
    if buddings_df is None:
        # mothers
        t_batch = [t for i in range(batch_size)]
        for idx in range(0, no_cells, batch_size):
            print(f" mother {idx:d}/{no_cells-1:d}")
            if idx + batch_size < no_cells:
                v_batch = [volume[idx + i, :] for i in range(batch_size)]
            else:
                v_batch = [volume[i, :] for i in range(idx, no_cells)]
                t_batch = [t for i in range(len(v_batch))]
            res_batch = pool.map(batch_GP, t_batch, v_batch)
            for ires in res_batch:
                if ires:
                    # store results as arrays
                    for s in signals:
                        res[s][idx, :] = ires[s]
    else:
        # buds
        for idx in range(no_cells):
            print(f" buds for {idx:d}/{no_cells-1:d}")
            ivol = volume[idx, :]
            # indices for buds
            bs = np.concatenate(([0], np.where(buddings[idx, :])[0], [t.size]))
            # get bud data
            t_batch = [t[start:end] for start, end in zip(bs, bs[1:])]
            v_batch = [ivol[start:end] for start, end in zip(bs, bs[1:])]
            res_batch = pool.map(batch_GP, t_batch, v_batch)
            for i, (start, end) in enumerate(zip(bs, bs[1:])):
                if res_batch[i]:
                    # store results as arrays
                    for s in signals:
                        res[s][idx, start:end] = res_batch[i][s]
    return res


def runGP(x, y, bounds, noruns, signals, verbose, title=None, maxnofigs=70):
    """
    Run GP on a single time series.

    Results returned as a dictionary.
    """
    # drop NaNs
    i_nan = np.isnan(y)
    x = x[~i_nan]
    y = y[~i_nan]
    # run GP
    if np.any(y):
        if verbose:
            print(title)
        # initiate GP
        mg = gp.maternGP(bounds, x, y)
        mg.findhyperparameters(noruns=noruns, exitearly=False)
        if verbose:
            # print hyperparameters
            mg.results()
            print()
        # predict smoothed data
        mg.predict(x, derivs=2)
        if verbose and len(plt.get_fignums()) < maxnofigs:
            plt.figure()
            plt.suptitle(title)
            plt.subplot(2, 1, 1)
            mg.sketch(".")
            plt.ylabel("volume")
            plt.subplot(2, 1, 2)
            mg.sketch(".", derivs=1)
            plt.ylabel("d/dt volume")
            plt.xlabel("time")
            plt.show(block=False)
        res = dict(
            zip(
                signals,
                reNaN(
                    [mg.f, mg.df, mg.ddf, mg.fvar, mg.dfvar, mg.ddfvar], i_nan
                ),
            )
        )
    else:
        # no data
        if verbose:
            print(title + "\n" + "data all NaN")
        res = None
    return res


def reNaN(oldlist, i_nan):
    """
    Add back NaN dropped from an array.

    Takes a list of arrays and puts back NaN for each array
    at indices i_nan.
    """
    if np.any(i_nan):
        newlist = []
        for x in oldlist:
            y = np.nan * np.ones(i_nan.size)
            y[~i_nan] = x
            newlist.append(y)
        return newlist
    else:
        return oldlist


def add_gr_from_pkl(dl, pkldir):
    """
    Load growth rate from pickle files.

    Add to and save dataframe in a dataloader instance.

    The following columns will be added to dataloader's data frame:

    smoothed_bud_volume : the bud volume smoothed by the Gaussian process
    bud_growth_rate : the time derivative of the smoothed bud volume
    d_bud_growth_rate : the second time-derivative of the smoothed bud volume
    smoothed_bud_volume_var : the square of the estimated error in smoothed_bud_volume
    bud_growth_rate_var :  the square of the estimated error
    d_bud_growth_rate_var: the square of the estimated error

    smoothed_volume : the mother's volume smoothed by the Gaussian process
    growth_rate : the time derivative of the mother's smoothed volume
    d_growth_rate: the second time derivative of the mother's smoothed volume
    smoothed_volume_var : the square of the estimated error in smoothed_volume
    growth_rate_var : the square of the estimated error
    d_growth_rate_var : the square of the estimated error

    Arguments
    ---------
    dl : dataloader instance
    pkldir: string
        The directory storing the pkl files.
    """
    with open(str(Path(pkldir) / (dl.dataname + "_res_m.pkl")), "rb") as file:
        res_m = pickle.load(file)
    with open(str(Path(pkldir) / (dl.dataname + "_res_b.pkl")), "rb") as file:
        res_b = pickle.load(file)
    # add to dataframe
    for res in [res_b, res_m]:
        for signal in res:
            tdf = long_df(res[signal], signal)
            dl.df = pd.merge(dl.df, tdf, on=["id", "time"], how="left")
