from pathlib import Path

import h5py

try:
    import napari
except ModuleNotFoundError:
    print(
        "Napari cannot be imported.\nRun",
        ' python -m pip install "napari[all]"',
    )
import numpy as np
from agora.io.cells import Cells
from aliby.io.image import dispatch_image
from aliby.io.omero import Dataset
from aliby.io.omero import UnsafeImage as OImage
from aliby.tile.tiler import Tiler


class ImageViewer:

    def __init__(self, h5file):
        self.h5file_path = Path(h5file)
        print(f"Viewing {str(self.h5file_path)}")
        self.full = {}

    @classmethod
    def local(cls, h5file: str, image_file: str):
        """
        View images from local files.

        Files are either zarr or organised in directories.
        """
        iv = cls(h5file)
        image_file_path = Path(image_file)
        if iv.h5file_path.exists() and image_file_path.exists():
            with dispatch_image(image_file_path)(image_file_path) as image:
                tiler = Tiler.from_h5(image, iv.h5file_path)
                cells = Cells.from_source(iv.h5file_path)
            iv.tiler = tiler
            iv.cells = cells
            return iv
        else:
            if not iv.h5file_path.exists():
                print(f" Trouble loading {h5file}.")
            if not image_file_path.exists():
                print(f" Trouble loading {image_file}.")

    @classmethod
    def remote(cls, h5file: str, server_info: dict, omero_id: int):
        """View images from OMERO."""
        iv = cls(h5file)
        with h5py.File(iv.h5file_path, "r") as f:
            # get image_id from the h5 file
            image_id = f.attrs.get("image_id")
        if image_id is None:
            # get image_id from OMERO
            with Dataset(omero_id, **server_info) as dataset_om:
                positions = dataset_om.get_position_ids()
                image_id = positions.get(iv.h5file_path.name.split(".")[0])
        if image_id is None:
            print("Can't find an image.")
        else:
            print(f"Using image ID {image_id}.")
            iv.image_id = image_id
            image = OImage(image_id, **server_info)
            print("Connected to OMERO.")
            iv.tiler = Tiler.from_h5(image, iv.h5file_path)
            iv.cells = Cells.from_source(iv.h5file_path)
            return iv

    def get_all_traps_with_cells(self, tpt_end, tpt_start=0, display=True):
        """List traps with cells."""
        cells = self.cells
        tpts = range(tpt_start, tpt_end)
        traps_with_cells = []
        for trap_id in cells.traps:
            for tpt in tpts:
                if tpt in cells.nonempty_tp_in_trap(trap_id):
                    traps_with_cells.append(trap_id)
                    break
        if display:
            print(f"Traps with cells {traps_with_cells}")
        return traps_with_cells

    def sample_traps_with_cells(self, no_cells, tpt_end, tpt_start=0):
        """Sample some traps that have cells."""
        traps_with_cells = self.get_all_traps_with_cells(
            tpt_end, tpt_start, display=False
        )
        rng = np.random.default_rng()
        samples = rng.integers(
            low=0,
            high=len(traps_with_cells),
            size=np.min([no_cells, len(traps_with_cells)]),
        )
        return samples

    def get_tiles(self, trap_id, tps, channels_to_skip=None, cell_only=True):
        """Get dict of tiles with channel indices as keys."""
        tiles_dict = {}
        if channels_to_skip is None:
            channels = self.tiler.channels
        else:
            channels = [
                ch for ch in self.tiler.channels if ch not in channels_to_skip
            ]
        channel_indices = [channels.index(ch) for ch in channels]
        for ch_index, ch in zip(channel_indices, channels):
            tile_dict_for_ch = self.get_all_tiles(tps, ch_index)
            tiles = [x[trap_id] for x in tile_dict_for_ch.values()]
            if ch == "Brightfield":
                tiles_dict[ch] = tiles
            else:
                masks = [
                    self.cells.at_time(tp, kind="mask").get(trap_id, [])
                    for tp in tps
                ]
                # some masks may be empty
                default_mask = [np.ones(self.cells.tile_size).astype(bool)]
                nmasks = [m if m else default_mask for m in masks]
                # combine all masks for each time point
                stacked_masks = [
                    np.stack([mask for mask in masks_tp]).max(axis=0)
                    for masks_tp in nmasks
                ]
                # make tiles with fluorescence only in mask pixels
                new_tiles = []
                for tile, stacked_mask in zip(tiles, stacked_masks):
                    tile[~stacked_mask] = 0
                    new_tiles.append(tile)
                tiles_dict[ch] = new_tiles
        return tiles_dict

    def get_outlines(self, trap_id, tps):
        """Get uniquely labelled outlines for each cell time point."""
        # get outlines for each time point
        outlines = [
            self.cells.at_time(tp, kind="edgemask").get(trap_id, [])
            for tp in tps
        ]
        # get cell labels for each time point
        cell_labels = [
            self.cells.labels_at_time(tp).get(trap_id, []) for tp in tps
        ]
        # generate one image with all cell outlines uniquely labelled per tile
        labelled_outlines = [
            (
                np.stack(
                    [
                        (outline * label)
                        for outline, label in zip(outlines_tp, labels_tp)
                    ]
                ).max(axis=0)
                if len(labels_tp)
                else np.zeros(self.cells.tile_size).astype(bool)
            )
            for outlines_tp, labels_tp in zip(outlines, cell_labels)
        ]
        return labelled_outlines

    def get_all_tiles(
        self,
        tps,
        channel_index,
        z=0,
    ):
        """
        Get dict with time points as keys and all available tiles as values.

        We assume only a single channel.
        """
        z = z or self.tiler.ref_z
        ch_tps = [(channel_index, tp) for tp in tps]
        for ch, tp in ch_tps:
            if (ch, tp) not in self.full:
                print(f"Getting {self.tiler.channels[ch]} at time point {tp}.")
                self.full[(ch, tp)] = self.tiler.get_tiles_timepoint(
                    tp, channels=[ch], z=[z]
                )[:, 0, 0, z, ...]
        tile_dict = {tp: self.full[(ch, tp)] for ch, tp in ch_tps}
        return tile_dict

    def get_data_for_viewing(self, trap_id, tps, channels_to_skip):
        """
        Get images and outlines as multidimensional arrays for Napari.

        Returns
        -------
        ts_images: array
            An array of images ordered by time points, channels, z stack,
            x and y values.
        ts_labels: array
            An array of images of cell outlines ordered by time points,
            z stack, x and y values.
        channels: list of str
            List of channel names.
        """
        # get outlines and tiles
        outlines = self.get_outlines(trap_id, tps)
        tiles_dict = self.get_tiles(trap_id, tps, channels_to_skip)
        channels = list(tiles_dict.keys())
        # put time series into one array with dimensions TCZYX
        ydim, xdim = tiles_dict[list(tiles_dict.keys())[0]][0].shape
        ts_images = np.zeros(
            (tps.size, len(tiles_dict), 1, ydim, xdim)
        ).astype(int)
        ts_labels = np.zeros((tps.size, 1, ydim, xdim)).astype(int)
        # make array of time series of tiles
        for ch_index, channel in enumerate(tiles_dict):
            for tp_index in range(tps.size):
                ts_images[tp_index, ch_index, 0, ...] = tiles_dict[channel][
                    tp_index
                ]
        # make array of time series of outlines with no channels dimension
        for tp_index in range(tps.size):
            ts_labels[tp_index, 0, ...] = outlines[tp_index]
        return ts_images, ts_labels, channels

    def concat(self, arrangement, image_dict, axis):
        """
        Concat dict of images into one image array.

        Following the vertical layout in arrangment.
        """
        # concatenate vertically into a list
        images_v = [
            np.concatenate(
                [image_dict[trap_id] for trap_id in v_stack], axis=axis
            )
            for v_stack in arrangement
        ]
        # concatenate horizontally into an array
        if len(arrangement) > 1:
            images = np.concatenate(
                [image_v_stack for image_v_stack in images_v], axis=axis + 1
            )
        else:
            images = images_v[0]
        return images

    def combine_tiles(self, ts_images_dict, ts_labels_dict, no_vertical_tiles):
        """Combine tiles into one image first vertically then horizontally."""
        no_tiles = len(ts_images_dict)
        trap_ids = list(ts_images_dict.keys())
        # find how tiles will be arranged in the concatenated image
        if no_tiles < no_vertical_tiles:
            arrangement = [trap_ids]
        else:
            arrangement = [
                trap_ids[i : min(i + no_vertical_tiles, no_tiles)]
                for i in range(
                    0,
                    int(
                        np.floor(no_tiles / no_vertical_tiles)
                        * no_vertical_tiles
                    ),
                    no_vertical_tiles,
                )
            ]
        if no_tiles > np.array(arrangement).size:
            print(
                f"Too many tiles: {no_tiles - np.array(arrangement).size}"
                " will be ignored."
            )
        # concatenate images
        print("Combining tiles...")
        ts_images = self.concat(arrangement, ts_images_dict, axis=3)
        ts_labels = self.concat(arrangement, ts_labels_dict, axis=2)
        return ts_images, ts_labels

    def view(
        self,
        trap_ids,
        tpt_end=10,
        tpt_start=0,
        channels_to_skip=None,
        no_vertical_tiles=3,
    ):
        """
        Use Napari to view all channels and outlines for a particular trap.

        Fluorescence channels will not be immediately visible.

        Parameters
        ----------
        trap_id: int
            The trap to be viewed.
        tps: int or array of ints
            Either the last time point to be viewed or a rage of time points
            to view.
            If None, all time points will be viewed, but gathering the images
            will be slow.
        """
        tps = np.arange(tpt_start, tpt_end + 1)
        if isinstance(trap_ids, int):
            trap_ids = [trap_ids]
        ts_images_dict, ts_labels_dict = {}, {}
        for trap_id in trap_ids:
            ts_images_dict[trap_id], ts_labels_dict[trap_id], channels = (
                self.get_data_for_viewing(trap_id, tps, channels_to_skip)
            )
        # combine tiles
        ts_images, ts_labels = self.combine_tiles(
            ts_images_dict, ts_labels_dict, no_vertical_tiles
        )
        # launch napari
        self.launch_napari(ts_images, ts_labels, channels)

    def launch_napari(self, ts_images, ts_labels, channels):
        """Use Napari to see the images and outlines."""
        viewer = napari.Viewer()
        viewer.add_image(
            ts_images[:, channels.index("Brightfield"), ...],
            name="Brightfield",
        )
        viewer.add_labels(ts_labels, name="outlines")
        # fluorescence channels are not initially visible
        for i, channel in enumerate(channels):
            if channel != "Brightfield":
                viewer.add_image(
                    ts_images[:, i, ...],
                    name=channel,
                    colormap=colormap(channel),
                    visible=False,
                    opacity=0.5,
                )


####
def colormap(channel):
    """Find default colormap."""
    if "GFP" in channel:
        colormap = "green"
    elif "Cherry" in channel or "RFP" in channel:
        colormap = "red"
    else:
        colormap = "gray"
    return colormap


def get_files(
    aliby_input: str,
    aliby_output: str,
    omero_name: str,
    position: str,
):
    """Find the h5 file and corresponding zarr file for one position."""
    h5files = [str(f) for f in (Path(aliby_output) / omero_name).glob("*.h5")]
    h5file = [f for f in h5files if position in f][0]
    image_file_name = h5file.split("/")[-1].split(".")[0] + ".zarr"
    image_file = str(Path(aliby_input) / omero_name / image_file_name)
    return {"h5file": h5file, "image_file": image_file}
