"""
Script to test how well we predict cytokinesis.

Requires data from a Myo1-GFP timelapse experiment for the ground truth.

Allows a grid search of the hyperparameters. There is a tradeoff between the number of predictions and their accuracy.
"""
import numpy as np
import matplotlib.pylab as plt

from mywela.dataloader import dataloader
from mywela.find_cytokinesis_diane import growth_rate_crossing
from mywela.find_cytokinesis import (
    find_cyto_pts_between_buddings,
    add_cyto_pts_from_Myo1,
    min_cell_cycle,
)
from mywela.growth_rate import addgrfrompkl
from mywela.plotting import plot_lineage


dl = dataloader(
    "/Users/pswain/ecdf/Swainlab/aliby_datasets/Alan",
    "../../tsv_data",
    ls=False,
)
# BY4741 Myo1-GFP Whi5-mCherry ; BY4741 Sfp1-GFP Nhp6A-mCherry
# switches every six hours

dataname = "19855_2021_03_20_sugarShift_pal_glu_pal_Myo1Whi5_Sfp1Nhp6a__00"
# dataname = "19856_2021_03_21_sugarShift_raf_pal_raf_Myo1Whi5_Sfp1Nhp6a__00"

# load data
if False:
    newfields = {
        "extraction/mCherry/np_max/max5px_med": "mCherry_max5px_med",
        "extraction/GFP/np_max/nuc_est_conv": "GFP_nuc_est_conv",
        "extraction/mCherry/np_max/nuc_est_conv": "mCherry_nuc_est_conv",
    }
    dl.load(dataname, g2a_dict={**dl.g2a_dict, **newfields})
    addgrfrompkl(dl, pkldir=".")
else:
    dl.load(dataname, use_tsv=True)

# prior: cytokinesis must be 20 minutes from each budding event
offset = int(20 / dl.dt)

# find peaks in Myo1 signal
myo1_signal = "GFP_nuc_est_conv"
myo1_column = "cyto_pts"
add_cyto_pts_from_Myo1(dl, myo1_signal, column_name=myo1_column)
# plot Myo1 for lineages and detected cyto_points
if False:
    for idx in dl.ids[:10]:
        plot_lineage(
            idx,
            dl.df,
            [myo1_signal],
            figsize=(10, 5),
            cyto_pts_signal=myo1_column,
        )


###
def corr(res):
    # within prior for time of cytokinesis
    withinprior = np.where(res[:, -1])
    # local definitions
    pred_cyt = res[:, 1][withinprior]
    true_cyt = res[:, 3][withinprior]
    corr = np.corrcoef(
        true_cyt[~np.isnan(pred_cyt)], pred_cyt[~np.isnan(pred_cyt)]
    )[0, 1]
    return corr, pred_cyt, true_cyt


###


# grid search for best threshold
no_bud_to_bud = 0
nofigs = 0
for thresh in [2]:
    res = []
    # predictions
    for idx in dl.ids:
        (
            myo1,
            t,
            bgr,
            bgrvar,
            mgr,
            mgrvar,
            buddings,
            true_cyto,
        ) = dl.get_lineage_data(
            idx,
            [
                "max5px_med",
                "time",
                "bud_growth_rate",
                "bud_growth_rate_var",
                "mother_growth_rate",
                "mother_growth_rate_var",
                "buddings",
                myo1_column,
            ],
        )
        # find one cyto_point between a pair of buddings
        bs = np.where(buddings)[0]
        for start, end in zip(bs, bs[1:]):
            if end > start + min_cell_cycle(dl.dt):
                no_bud_to_bud += 1
                pks = np.where(true_cyto[start:end])[0]
                # consider only time series with one Myo1 peak
                if np.any(pks) and pks.size == 1:
                    ltpk = pks[0]
                    tpk = ltpk + start
                    # make prediction
                    pred = find_cyto_pts_between_buddings(
                        bgr[start:end],
                        bgrvar[start:end],
                        mgr[start:end],
                        mgrvar[start:end],
                        offs=offset,
                        thresh=thresh,
                    )
                    # store results
                    lres = []
                    if np.isnan(pred):
                        # absolute time and time wrt budding event
                        lres.append([np.nan, np.nan])
                    else:
                        # absolute time and time wrt budding event
                        lres.append(
                            [t[start + pred], t[start + pred] - t[start]]
                        )
                    # store true cytokinesis point
                    if tpk < start + offset or tpk > end - offset:
                        # without prior on time of cytokinesis
                        # absolute time; time wrt budding event; whether within prior
                        lres.append([t[tpk], t[tpk] - t[start], False])
                    else:
                        # within prior
                        lres.append([t[tpk], t[tpk] - t[start], True])
                    lres = [lr for lrlist in lres for lr in lrlist]
                    res.append(lres)
                    if ~np.isnan(pred) and nofigs < 20:
                        plt.figure()
                        lt = np.arange(end - start)
                        norm = np.nanmean(mgr[start:end])
                        plt.plot(
                            lt,
                            bgr[start:end] / norm,
                            lt,
                            mgr[start:end] / norm,
                        )
                        plt.fill_between(
                            lt,
                            bgr[start:end] / norm
                            - np.sqrt(bgrvar[start:end]) / norm,
                            bgr[start:end] / norm
                            + np.sqrt(bgrvar[start:end]) / norm,
                            alpha=0.3,
                        )
                        plt.fill_between(
                            lt,
                            mgr[start:end] / norm
                            - np.sqrt(mgrvar[start:end]) / norm,
                            mgr[start:end] / norm
                            + np.sqrt(mgrvar[start:end]) / norm,
                            alpha=0.3,
                        )
                        plt.plot(
                            ltpk,
                            mgr[start:end][ltpk] / norm,
                            "rx",
                            label="true",
                        )
                        plt.plot(
                            pred,
                            mgr[start:end][pred] / norm,
                            "go",
                            label="pred",
                        )
                        plt.title(f"{idx:s} -- {start:d}, {end:d}")
                        plt.grid()
                        plt.legend()
                        plt.show()
                        nofigs += 1
    res = np.array(res)
    corr, pred_cyt, true_cyt = corr(res)
    # scores for cytokinesis points within prior
    sc = np.log2(pred_cyt / true_cyt)
    print()
    print(f"thresh = {thresh:.2f} ; no_bud_to_bud = {no_bud_to_bud:4d}")
    print(
        f"sc : {np.nanmean(sc):.2e} +/- {np.nanstd(sc):.2f} ; "
        f"abs(sc) : {np.nanmean(np.abs(sc)):.3f} ; "
        f"no predictions : {sc[~np.isnan(sc)].size:d} ; ",
        f"corr : {corr:.3f}",
    )

    # histogram of scores for cytokinesis points within prior
    plt.figure()
    plt.hist(sc, bins=20, alpha=0.5, density=True)
    plt.title("score")
    plt.show()
    # correlation between true and predicted
    plt.figure()
    plt.plot(true_cyt, pred_cyt, ".")
    plt.plot(true_cyt, true_cyt, "k-")
    plt.xlabel("true")
    plt.ylabel("predicted")
    plt.xscale("log")
    plt.yscale("log")
    plt.title("corr = {:.2f}".format(corr))
    plt.show()
